@@ -1,6 +1,7 @@
 import time
 import warnings
 from typing import List
+from typing import Literal
 from typing import Optional
 from typing import Union
 from typing import cast
@@ -12,9 +13,9 @@ import statsmodels.api as sm  # type: ignore
 from joblib import Parallel  # type: ignore
 from joblib import delayed
 from joblib import parallel_backend
+from scipy.optimize import minimize
 from scipy.special import polygamma  # type: ignore
 from scipy.stats import f  # type: ignore
-from scipy.stats import norm
 from statsmodels.tools.sm_exceptions import DomainWarning  # type: ignore
 
 from pydeseq2.preprocessing import deseq2_norm
@@ -26,6 +27,8 @@ from pydeseq2.utils import fit_moments_dispersions
 from pydeseq2.utils import fit_rough_dispersions
 from pydeseq2.utils import get_num_processes
 from pydeseq2.utils import irls_solver
+from pydeseq2.utils import mean_absolute_deviation
+from pydeseq2.utils import nb_nll
 from pydeseq2.utils import robust_method_of_moments_disp
 from pydeseq2.utils import test_valid_counts
 from pydeseq2.utils import trimmed_mean
@@ -35,7 +38,8 @@ warnings.simplefilter("ignore", DomainWarning)
 # Ignore AnnData's FutureWarning about implicit data conversion.
 warnings.simplefilter("ignore", FutureWarning)
 
-# TODO: support loading / saving DeseqDataSets from AnnData objects.
+
+# TODO: implement a quiet (non-verbose) mode ?
 
 
 class DeseqDataSet(ad.AnnData):
@@ -50,6 +54,11 @@ class DeseqDataSet(ad.AnnData):
 
     Parameters
     ----------
+    adata : anndata.AnnData
+        AnnData from which to initialize the DeseqDataSet. Must have counts ('X') and
+        clinical metadata ('obs') fields. If ``None``, both ``counts`` and ``clinical``
+        arguments must be provided.
+
     counts : pandas.DataFrame
         Raw counts. One column per gene, rows are indexed by sample barcodes.
 
@@ -58,14 +67,13 @@ class DeseqDataSet(ad.AnnData):
         Must be indexed by sample barcodes.
 
     design_factors : str or list
-        Name of the columns of clinical to be used as design variables. If a list,
-        the last factor will be considered the variable of interest by default.
+        Name of the columns of clinical to be used as design variables.
         Only bi-level factors are supported. (default: ``'condition'``).
 
-    reference_level : str
-        The factor to use as a reference. Must be one of the values taken by the design.
-        If None, the reference will be chosen alphabetically (last in order).
-        (default: ``None``).
+    tested_level : list or None
+        An optional list of two strings of the form ``["factor", "test_level"]``
+        specifying the factor of interest and the (non-control) level we're testing, e.g.
+        ``["condition", "B"]``. (default: ``None``).
 
     min_mu : float
         Threshold for mean estimates. (default: ``0.5``).
@@ -151,10 +159,12 @@ class DeseqDataSet(ad.AnnData):
 
     def __init__(
         self,
-        counts: pd.DataFrame,
-        clinical: pd.DataFrame,
+        *,
+        adata: Optional[ad.AnnData] = None,
+        counts: Optional[pd.DataFrame] = None,
+        clinical: Optional[pd.DataFrame] = None,
         design_factors: Union[str, List[str]] = "condition",
-        reference_level: Optional[str] = None,
+        tested_level: Optional[List[str]] = None,
         min_mu: float = 0.5,
         min_disp: float = 1e-8,
         max_disp: float = 10.0,
@@ -166,11 +176,21 @@ class DeseqDataSet(ad.AnnData):
         joblib_verbosity: int = 0,
     ) -> None:
 
-        # Test counts before going further
-        test_valid_counts(counts)
+        if adata is None and (counts is None or clinical is None):
+            raise ValueError(
+                "Either adata or both counts and clinical arguments must be provided."
+            )
 
         # Initialize the AnnData part
-        super().__init__(X=counts, obs=clinical, dtype=int)
+        if adata is not None:
+            # Test counts before going further
+            test_valid_counts(adata.X)
+            # Copy fields from original AnnData
+            self.__dict__.update(adata.__dict__)
+        else:
+            # Test counts before going further
+            test_valid_counts(counts)
+            super().__init__(X=counts, obs=clinical, dtype=int)
 
         # Convert design_factors to list if a single string was provided.
         self.design_factors = (
@@ -185,7 +205,7 @@ class DeseqDataSet(ad.AnnData):
         self.obsm["design_matrix"] = build_design_matrix(
             clinical_df=self.obs,
             design_factors=self.design_factors,
-            ref=reference_level,
+            tested_level=tested_level,
             expanded=False,
             intercept=True,
         )
@@ -200,6 +220,50 @@ class DeseqDataSet(ad.AnnData):
         self.batch_size = batch_size
         self.joblib_verbosity = joblib_verbosity
 
+    def vst(self, use_design: bool = False) -> None:
+        """
+        Fit a variance stabilizing transformation, and apply it to normalized counts.
+
+        Results are stored in ``dds.layers["vst_counts"]``.
+
+        Parameters
+        ----------
+        use_design : bool
+            Whether to use the full design matrix to fit dispersions and the trend curve.
+            If False, only an intercept is used. (default: ``False``).
+        """
+
+        # Start by fitting median-of-ratio size factors, if not already present.
+        if "size_factors" not in self.obsm:
+            self.fit_size_factors()
+
+        if use_design:
+            # Check that the dispersion trend curve was fitted. If not, fit it.
+            # This will call previous functions in a cascade.
+            if "trend_coeffs" not in self.uns:
+                self.fit_dispersion_trend()
+        else:
+            # Reduce the design matrix to an intercept and reconstruct at the end
+            self.obsm["design_matrix_buffer"] = self.obsm["design_matrix"].copy()
+            self.obsm["design_matrix"] = pd.DataFrame(
+                1, index=self.obs_names, columns=[["intercept"]]
+            )
+            # Fit the trend curve with an intercept design
+            self.fit_genewise_dispersions()
+            self.fit_dispersion_trend()
+
+            # Restore the design matrix and free buffer
+            self.obsm["design_matrix"] = self.obsm["design_matrix_buffer"].copy()
+            del self.obsm["design_matrix_buffer"]
+
+        # Apply VST
+        a0, a1 = self.uns["trend_coeffs"]
+        cts = self.layers["normed_counts"]
+        self.layers["vst_counts"] = np.log2(
+            (1 + a1 + 2 * a0 * cts + 2 * np.sqrt(a0 * cts * (1 + a1 + a0 * cts)))
+            / (4 * a0)
+        )
+
     def deseq2(self) -> None:
         """Perform dispersion and log fold-change (LFC) estimation.
 
@@ -227,14 +291,36 @@ class DeseqDataSet(ad.AnnData):
             # for genes that had outliers replaced
             self.refit()
 
-    def fit_size_factors(self) -> None:
+    def fit_size_factors(
+        self, fit_type: Literal["ratio", "iterative"] = "ratio"
+    ) -> None:
         """Fit sample-wise deseq2 normalization (size) factors.
 
-        Uses the median-of-ratios method: see :func:`pydeseq2.preprocessing.deseq2_norm`.
+        Uses the median-of-ratios method: see :func:`pydeseq2.preprocessing.deseq2_norm`,
+        unless each gene has at least one sample with zero read counts, in which case it
+        switches to the ``iterative`` method.
+
+        Parameters
+        ----------
+        fit_type : str
+            The normalization method to use (default: ``"ratio"``).
         """
         print("Fitting size factors...")
         start = time.time()
-        self.layers["normed_counts"], self.obsm["size_factors"] = deseq2_norm(self.X)
+        if fit_type == "iterative":
+            self._fit_iterate_size_factors()
+        # Test whether it is possible to use median-of-ratios.
+        elif (self.X == 0).any(0).all():
+            # There is at least a zero for each gene
+            warnings.warn(
+                "Every gene contains at least one zero, "
+                "cannot compute log geometric means. Switching to iterative mode.",
+                RuntimeWarning,
+                stacklevel=2,
+            )
+            self._fit_iterate_size_factors()
+        else:
+            self.layers["normed_counts"], self.obsm["size_factors"] = deseq2_norm(self.X)
         end = time.time()
         print(f"... done in {end - start:.2f} seconds.\n")
 
@@ -423,7 +509,7 @@ class DeseqDataSet(ad.AnnData):
         """
 
         # Check that the dispersion trend curve was fitted. If not, fit it.
-        if "trend_coeffs" not in self.uns:
+        if "fitted_dispersions" not in self.varm:
             self.fit_dispersion_trend()
 
         # Exclude genes with all zeroes
@@ -441,9 +527,9 @@ class DeseqDataSet(ad.AnnData):
             100 * self.min_disp
         )
 
-        self.uns["_squared_logres"] = np.median(
-            np.abs(disp_residuals[above_min_disp])
-        ) ** 2 / norm.ppf(0.75)
+        self.uns["_squared_logres"] = (
+            mean_absolute_deviation(disp_residuals[above_min_disp]) ** 2
+        )
         self.uns["prior_disp_var"] = np.maximum(
             self.uns["_squared_logres"] - polygamma(1, (num_genes - num_vars) / 2),
             0.25,
@@ -498,7 +584,7 @@ class DeseqDataSet(ad.AnnData):
 
         # Filter outlier genes for which we won't apply shrinkage
         self.varm["dispersions"] = self.varm["MAP_dispersions"].copy()
-        self.varm["_outlier_genes"] = np.log(self.varm["dispersions"]) > np.log(
+        self.varm["_outlier_genes"] = np.log(self.varm["genewise_dispersions"]) > np.log(
             self.varm["fitted_dispersions"]
         ) + 2 * np.sqrt(self.uns["_squared_logres"])
         self.varm["dispersions"][self.varm["_outlier_genes"]] = self.varm[
@@ -793,3 +879,82 @@ class DeseqDataSet(ad.AnnData):
             new_all_zeroes.sum()
         )
         self[:, self.new_all_zeroes_genes].varm["LFC"] = np.zeros(new_all_zeroes.sum())
+
+    def _fit_iterate_size_factors(self, niter: int = 10, quant: float = 0.95) -> None:
+        """
+        Fit size factors using the ``iterative`` method.
+
+        Used when each gene has at least one zero.
+
+        Parameters
+        ----------
+        niter : int
+            Maximum number of iterations to perform (default: ``10``).
+
+        quant : float
+            Quantile value at which negative likelihood is cut in the optimization
+            (default: ``0.95``).
+
+        """
+
+        self.obsm["size_factors"] = np.ones(self.n_obs)
+
+        # Reduce the design matrix to an intercept and reconstruct at the end
+        self.obsm["design_matrix_buffer"] = self.obsm["design_matrix"].copy()
+        self.obsm["design_matrix"] = pd.DataFrame(
+            1, index=self.obs_names, columns=[["intercept"]]
+        )
+
+        # Fit size factors using MLE
+        def objective(p):
+            sf = np.exp(p - np.mean(p))
+            nll = nb_nll(
+                counts=self[:, self.non_zero_genes].X,
+                mu=self[:, self.non_zero_genes].layers["_mu_hat"]
+                / self.obsm["size_factors"][:, None]
+                * sf[:, None],
+                alpha=self[:, self.non_zero_genes].varm["dispersions"],
+            )
+            # Take out the lowest likelihoods (highest neg) from the sum
+            return np.sum(nll[nll < np.quantile(nll, quant)])
+
+        for i in range(niter):
+            # Estimate dispersions based on current size factors
+            self.fit_genewise_dispersions()
+
+            # Use a mean trend curve
+            use_for_mean_genes = self.var_names[
+                (self.varm["genewise_dispersions"] > 10 * self.min_disp)
+                & self.varm["non_zero"]
+            ]
+
+            mean_disp = trimmed_mean(
+                self[:, use_for_mean_genes].varm["genewise_dispersions"], trim=0.001
+            )
+            self.varm["fitted_dispersions"] = np.ones(self.n_vars) * mean_disp
+            self.fit_dispersion_prior()
+            self.fit_MAP_dispersions()
+            old_sf = self.obsm["size_factors"].copy()
+
+            # Fit size factors using MLE
+            res = minimize(objective, np.log(old_sf), method="Powell")
+
+            self.obsm["size_factors"] = np.exp(res.x - np.mean(res.x))
+
+            if not res.success:
+                print("A size factor fitting iteration failed.")
+                break
+
+            if (i > 1) and np.sum(
+                (np.log(old_sf) - np.log(self.obsm["size_factors"])) ** 2
+            ) < 1e-4:
+                break
+            elif i == niter - 1:
+                print("Iterative size factor fitting did not converge.")
+
+        # Restore the design matrix and free buffer
+        self.obsm["design_matrix"] = self.obsm["design_matrix_buffer"].copy()
+        del self.obsm["design_matrix_buffer"]
+
+        # Store normalized counts
+        self.layers["normed_counts"] = self.X / self.obsm["size_factors"][:, None]
