@@ -1,7 +1,10 @@
+from typing import Tuple
+
 import numpy as np
+import pandas as pd
 
 
-def deseq2_norm(counts):
+def deseq2_norm(counts: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
     """
     Return normalized counts and size_factors.
 
@@ -9,8 +12,8 @@ def deseq2_norm(counts):
 
     Parameters
     ----------
-    counts : pandas.DataFrame
-            Raw counts. One column per gene, rows are indexed by sample barcodes.
+    counts : ndarray
+            Raw counts. One column per gene, one row per sample.
 
     Returns
     -------
@@ -21,17 +24,19 @@ def deseq2_norm(counts):
     size_factors : pandas.DataFrame
         DESeq2 normalization factors.
     """
+
     # Compute gene-wise mean log counts
-    log_counts = counts.apply(np.log)
+    with np.errstate(divide="ignore"):  # ignore division by zero warnings
+        log_counts = np.log(counts)
     logmeans = log_counts.mean(0)
     # Filter out genes with -∞ log means
-    filtered_genes = ~np.isinf(logmeans).values
+    filtered_genes = ~np.isinf(logmeans)
     # Subtract filtered log means from log counts
-    log_ratios = log_counts.iloc[:, filtered_genes] - logmeans[filtered_genes]
+    log_ratios = log_counts[:, filtered_genes] - logmeans[filtered_genes]
     # Compute sample-wise median of log ratios
-    log_medians = log_ratios.median(1)
+    log_medians = np.median(log_ratios, axis=1)
     # Return raw counts divided by size factors (exponential of log ratios)
     # and size factors
     size_factors = np.exp(log_medians)
-    deseq2_counts = counts.div(size_factors, 0)
+    deseq2_counts = counts / size_factors[:, None]
     return deseq2_counts, size_factors
